import warnings
from safeserialize import write, read
from ..core import writer as core_writer, reader as core_reader
from .numpy import _allowed_dtypes as _numpy_dtypes

VERSION = 1

_pandas_dtypes = {
    "boolean",
    "Int8", "Int16", "Int32", "Int64",
    "UInt8", "UInt16", "UInt32", "UInt64",
    "Float32", "Float64",
}

_WARNING_SHOWN = False

def _warn_experimental():
    global _WARNING_SHOWN
    if _WARNING_SHOWN:
        return

    _WARNING_SHOWN = True

    warning_message = (
        "Serialization of Pandas objects is still experimental. "
        "The binary format can change at any time. "
        "Please verify that loads(dumps(data)) returns your original data "
        "and report any bugs you might encounter: "
        "https://github.com/99991/safeserialize/issues"
    )
    warnings.warn(warning_message, UserWarning, stacklevel=5)

def writer(type_str):
    original_writer = core_writer(type_str)
    def decorator(func):
        def wrapper(*args, **kwargs):
            _warn_experimental()
            return func(*args, **kwargs)
        return original_writer(wrapper)
    return decorator

def reader(type_str):
    original_reader = core_reader(type_str)
    def decorator(func):
        def wrapper(*args, **kwargs):
            _warn_experimental()
            return func(*args, **kwargs)
        return original_reader(wrapper)
    return decorator

@writer("pandas._libs.missing.NAType")
def write_na_type(data, out):
    write(VERSION, out)

@reader("pandas._libs.missing.NAType")
def read_na_type(f):
    version = read(f)
    assert version == VERSION
    import pandas as pd
    return pd.NA

@writer("pandas.core.indexes.range.RangeIndex")
def write_range_index(index, out):
    write(index.start, out)
    write(index.stop, out)
    write(index.step, out)

@reader("pandas.core.indexes.range.RangeIndex")
def read_range_index(f):
    start = read(f)
    stop = read(f)
    step = read(f)
    import pandas as pd
    return pd.RangeIndex(start, stop, step)

@writer("pandas.core.indexes.frozen.FrozenList")
def writer_frozen_list(data, out):
    write(list(data), out)

@reader("pandas.core.indexes.frozen.FrozenList")
def reader_frozen_list(f):
    data = read(f)
    import pandas
    return pandas.core.indexes.frozen.FrozenList(data)

@writer("pandas.core.indexes.base.Index")
def write_base_index(index, out):
    dtype = index.dtype
    dtype_name = dtype.name

    assert dtype_name in _numpy_dtypes
    write(index.name, out)
    write(dtype_name, out)
    write(index.names, out)
    write(index._data, out)

@reader("pandas.core.indexes.base.Index")
def reader_base_index(f):
    import pandas as pd

    name = read(f)
    dtype_name = read(f)
    assert dtype_name in _numpy_dtypes
    names = read(f)
    data = read(f)

    index = pd.Index(data, dtype=dtype_name, name=name)
    index.names = names
    return index

@writer("pandas.core.series.Series")
def write_series(series, out):
    import numpy
    import pandas

    values = series.values
    values_dtype_name = values.dtype.name

    write(VERSION, out)
    write(series.name, out)
    write(series.dtype, out)
    write(series.values.dtype, out)
    write(series.index, out)

    if values_dtype_name == "string":
        assert isinstance(values, pandas.core.arrays.string_.StringArray)
        write(values.tolist(), out)

    elif values_dtype_name in _pandas_dtypes:
        write(values.isna(), out)
        values_numpy = values._data
        assert isinstance(values_numpy, numpy.ndarray)
        write(values_numpy, out)

    elif values_dtype_name in _numpy_dtypes:
        assert isinstance(values, numpy.ndarray)
        write(values, out)

    elif values_dtype_name == "category":
        write(values.categories, out)
        assert isinstance(values.codes, numpy.ndarray)
        write(values.codes, out)
        write(values.ordered, out)

    else:
        raise ValueError(f"Pandas dtype {values_dtype_name} not implemented")

@reader("pandas.core.series.Series")
def read_series(f):
    import pandas as pd
    import numpy as np

    version = read(f)
    assert version == VERSION
    series_name = read(f)
    series_dtype = read(f)
    values_dtype = read(f)
    values_dtype_name = values_dtype.name
    index = read(f)

    if values_dtype_name == "string":
        values = read(f)
        array = pd.array(values, dtype="string")
        series = pd.Series(array, dtype="string", index=index)

    elif values_dtype_name in _numpy_dtypes:
        values = read(f)

        # NumPy datetime64[ns] does not have timezone information.
        # But pd.Series does, so if the series_dtype contains a timezone,
        # we have to make sure that we remove that and apply it later
        # or else pd.Series will change our times to account for the
        # timezone difference between NumPy (UTC by default) and pandas.
        if isinstance(series_dtype, pd.DatetimeTZDtype):
            # Create series from timezone-less dtype
            series = pd.Series(values, dtype=series_dtype.base, index=index)
            # and apply actual dtype afterwards
            series = series.dt.tz_localize("UTC")
            series = series.dt.tz_convert(series_dtype.tz)
        else:
            series = pd.Series(values, dtype=series_dtype, index=index)

    elif values_dtype_name in _pandas_dtypes:
        isna = read(f)
        assert isna.dtype == np.bool_
        values = read(f)
        series = pd.Series(values, dtype=series_dtype, index=index)
        series = series.mask(isna)

    elif values_dtype_name == "category":
        categories = read(f)
        codes = read(f)
        # `ordered` is unused, already stored in categories
        ordered = read(f)
        assert isinstance(ordered, bool)
        categorical = pd.Categorical.from_codes(codes, categories)
        series = pd.Series(categorical, dtype=series_dtype, index=index)

    else:
        raise ValueError(f"Pandas dtype {dtype_name} not implemented")

    series.name = series_name

    return series

@writer("pandas.core.frame.DataFrame")
def write_dataframe(data, out):
    write(VERSION, out)

    m, n = data.shape
    write(m, out)
    write(n, out)
    write(data.index, out)

    for _, series in data.items():
        write(series, out)

@reader("pandas.core.frame.DataFrame")
def read_dataframe(f):
    import pandas as pd

    version = read(f)
    assert version == VERSION

    m = read(f)
    n = read(f)
    index = read(f)

    series = [read(f) for _ in range(n)]

    df = pd.concat(series, axis=1)
    df.index = index

    assert df.shape == (m, n)

    return df

pandas_dtypes = [
    ("pandas.core.arrays.integer.Int8Dtype", "Int8Dtype"),
    ("pandas.core.arrays.integer.Int16Dtype", "Int16Dtype"),
    ("pandas.core.arrays.integer.Int32Dtype", "Int32Dtype"),
    ("pandas.core.arrays.integer.Int64Dtype", "Int64Dtype"),
    ("pandas.core.arrays.integer.UInt8Dtype", "UInt8Dtype"),
    ("pandas.core.arrays.integer.UInt16Dtype", "UInt16Dtype"),
    ("pandas.core.arrays.integer.UInt32Dtype", "UInt32Dtype"),
    ("pandas.core.arrays.integer.UInt64Dtype", "UInt64Dtype"),
    ("pandas.core.arrays.floating.Float32Dtype", "Float32Dtype"),
    ("pandas.core.arrays.floating.Float64Dtype", "Float64Dtype"),
    ("pandas.core.arrays.boolean.BooleanDtype", "BooleanDtype"),
    ("pandas.core.arrays.string_.StringDtype", "StringDtype"),
]

for dtype_path, dtype_name in pandas_dtypes:
    def make_dtype_reader_writer(dtype_path, dtype_name):
        @writer(dtype_path)
        def writer_func(data, out):
            pass

        @reader(dtype_path)
        def reader_func(f):
            import pandas as pd
            return getattr(pd, dtype_name)()

    make_dtype_reader_writer(dtype_path, dtype_name)

@writer("pandas.core.dtypes.dtypes.CategoricalDtype")
def write_CategoricalDtype(data, out):
    import pandas
    assert isinstance(data.categories, pandas.core.indexes.base.Index)
    write(data.categories, out)
    write(data.ordered, out)

@reader("pandas.core.dtypes.dtypes.CategoricalDtype")
def read_CategoricalDtype(f):
    import pandas
    categories = read(f)
    ordered = read(f)
    return pandas.CategoricalDtype(categories, ordered=ordered)

@writer("pandas.core.dtypes.dtypes.DatetimeTZDtype")
def write_DatetimeTZDtype(data, out):
    write(data.unit, out)
    write(data.tz, out)

@reader("pandas.core.dtypes.dtypes.DatetimeTZDtype")
def read_DatetimeTZDtype(f):
    unit = read(f)
    tz = read(f)
    import pandas
    return pandas.DatetimeTZDtype(unit, tz)

@writer("pandas._libs.tslibs.timedeltas.Timedelta")
def write_Timedelta(data, out):
    write(data.value, out)
    write(data.unit, out)

@reader("pandas._libs.tslibs.timedeltas.Timedelta")
def read_Timedelta(f):
    value = read(f)
    unit = read(f)
    import pandas
    return pandas.Timedelta(value, unit=unit)

@writer("pandas.core.indexes.datetimes.DatetimeIndex")
def write_DatetimeIndex(index, out):
    write(index.values, out)
    write(index.tz, out)
    write(index.name, out)

@reader("pandas.core.indexes.datetimes.DatetimeIndex")
def read_DatetimeIndex(f):
    import pandas as pd
    values = read(f)
    tz = read(f)
    name = read(f)
    values = pd.Series(values, name=name)
    values = values.dt.tz_localize("UTC").dt.tz_convert(tz)
    return pd.DatetimeIndex(values)

@writer("pandas.core.indexes.timedeltas.TimedeltaIndex")
def write_TimedeltaIndex(index, out):
    write(index.values, out)
    write(index.name, out)
    write(index.freqstr, out)

@reader("pandas.core.indexes.timedeltas.TimedeltaIndex")
def read_TimedeltaIndex(f):
    import pandas as pd
    values = read(f)
    name = read(f)
    freq = read(f)
    return pd.TimedeltaIndex(values, name=name, freq=freq)

@writer("pandas.core.indexes.category.CategoricalIndex")
def write_CategoricalIndex(index, out):
    write(index.codes, out)
    write(index.categories, out)
    write(index.ordered, out)
    write(index.name, out)

@reader("pandas.core.indexes.category.CategoricalIndex")
def read_CategoricalIndex(f):
    codes = read(f)
    categories = read(f)
    ordered = read(f)
    name = read(f)
    import pandas as pd
    cat = pd.Categorical.from_codes(codes, categories, ordered=ordered)
    return pd.CategoricalIndex(cat, name=name)

@writer("pandas.core.indexes.interval.IntervalIndex")
def write_interval_index(index, out):
    write(index.left, out)
    write(index.right, out)
    write(index.closed, out)
    write(index.name, out)

@reader("pandas.core.indexes.interval.IntervalIndex")
def read_interval_index(f):
    left = read(f)
    right = read(f)
    closed = read(f)
    name = read(f)
    import pandas as pd
    return pd.IntervalIndex.from_arrays(
        left=left,
        right=right,
        closed=closed,
        name=name)

@writer("pandas.core.indexes.multi.MultiIndex")
def write_MultiIndex(index, out):
    write(index.levels, out)
    write(index.codes, out)
    write(index.names, out)
    write(index.sortorder, out)

@reader("pandas.core.indexes.multi.MultiIndex")
def read_MultiIndex(f):
    import pandas as pd
    levels = read(f)
    codes = read(f)
    names = read(f)
    sortorder = read(f)
    return pd.MultiIndex(
        levels=levels,
        codes=codes,
        names=names,
        sortorder=sortorder)

@writer("pandas.core.indexes.period.PeriodIndex")
def write_PeriodIndex(index, out):
    write(index.name, out)
    write(index.freqstr, out)
    write(index.asi8, out)

@reader("pandas.core.indexes.period.PeriodIndex")
def read_PeriodIndex(f):
    import pandas as pd
    name = read(f)
    freq = read(f)
    ordinals = read(f)
    return pd.PeriodIndex.from_ordinals(ordinals, freq=freq, name=name)



VERSION = 1
# Adapted from PNG file signature
# https://www.libpng.org/pub/png/spec/1.2/PNG-Rationale.html#R.PNG-file-signature
FILE_SIGNATURE = b'\x89SER\r\n\x1a\n'

TYPE_BOOL_FALSE = 0
TYPE_BOOL_TRUE = 1
TYPE_INT1 = 2
TYPE_INT2 = 3
TYPE_INT4 = 4
TYPE_INT8 = 5
TYPE_LARGE_INT = 6
# 48 = ord('0')
TYPE_INT_VALUE_0 = 48

builtins = {
    # String/bytes
    "builtins.str": 20,
    "builtins.bytes": 21,
    "builtins.bytearray": 22,

    # Collections
    "builtins.list": 23,
    "builtins.tuple": 24,
    "builtins.dict": 25,
    "builtins.set": 26,
    "builtins.frozenset": 27,

    # Scalar
    "builtins.NoneType": 28,
    "builtins.bool": 29,
    "builtins.float": 30,
    "builtins.complex": 31,
    "builtins.ellipsis": 32,
    "builtins.NotImplementedType": 33,

    # Misc
    "builtins.range": 34,
    "builtins.slice": 35,
}

TYPE_CUSTOM = 255



import re
import os

def test_readme():
    with open("README.md") as f:
        text = f.read()

    # Find code blocks
    pattern = r"```python\n(.*?)```"
    code_blocks = re.findall(pattern, text, re.DOTALL)

    for code in code_blocks:
        exec(code)

    os.remove("data.safeserialize")



from decimal import Decimal

from ...core import writer, reader, read_str, write_str

@writer("decimal.Decimal")
def write_decimal(data, out):
    write_str(str(data), out)

@reader("decimal.Decimal")
def read_decimal(f):
    return Decimal(read_str(f))



from ..core import writer, reader

_timezones = """
Africa/Abidjan
Africa/Accra
Africa/Addis_Ababa
Africa/Algiers
Africa/Asmara
Africa/Asmera
Africa/Bamako
Africa/Bangui
Africa/Banjul
Africa/Bissau
Africa/Blantyre
Africa/Brazzaville
Africa/Bujumbura
Africa/Cairo
Africa/Casablanca
Africa/Ceuta
Africa/Conakry
Africa/Dakar
Africa/Dar_es_Salaam
Africa/Djibouti
Africa/Douala
Africa/El_Aaiun
Africa/Freetown
Africa/Gaborone
Africa/Harare
Africa/Johannesburg
Africa/Juba
Africa/Kampala
Africa/Khartoum
Africa/Kigali
Africa/Kinshasa
Africa/Lagos
Africa/Libreville
Africa/Lome
Africa/Luanda
Africa/Lubumbashi
Africa/Lusaka
Africa/Malabo
Africa/Maputo
Africa/Maseru
Africa/Mbabane
Africa/Mogadishu
Africa/Monrovia
Africa/Nairobi
Africa/Ndjamena
Africa/Niamey
Africa/Nouakchott
Africa/Ouagadougou
Africa/Porto-Novo
Africa/Sao_Tome
Africa/Timbuktu
Africa/Tripoli
Africa/Tunis
Africa/Windhoek
America/Adak
America/Anchorage
America/Anguilla
America/Antigua
America/Araguaina
America/Argentina/Buenos_Aires
America/Argentina/Catamarca
America/Argentina/ComodRivadavia
America/Argentina/Cordoba
America/Argentina/Jujuy
America/Argentina/La_Rioja
America/Argentina/Mendoza
America/Argentina/Rio_Gallegos
America/Argentina/Salta
America/Argentina/San_Juan
America/Argentina/San_Luis
America/Argentina/Tucuman
America/Argentina/Ushuaia
America/Aruba
America/Asuncion
America/Atikokan
America/Atka
America/Bahia
America/Bahia_Banderas
America/Barbados
America/Belem
America/Belize
America/Blanc-Sablon
America/Boa_Vista
America/Bogota
America/Boise
America/Buenos_Aires
America/Cambridge_Bay
America/Campo_Grande
America/Cancun
America/Caracas
America/Catamarca
America/Cayenne
America/Cayman
America/Chicago
America/Chihuahua
America/Ciudad_Juarez
America/Coral_Harbour
America/Cordoba
America/Costa_Rica
America/Coyhaique
America/Creston
America/Cuiaba
America/Curacao
America/Danmarkshavn
America/Dawson
America/Dawson_Creek
America/Denver
America/Detroit
America/Dominica
America/Edmonton
America/Eirunepe
America/El_Salvador
America/Ensenada
America/Fort_Nelson
America/Fort_Wayne
America/Fortaleza
America/Glace_Bay
America/Godthab
America/Goose_Bay
America/Grand_Turk
America/Grenada
America/Guadeloupe
America/Guatemala
America/Guayaquil
America/Guyana
America/Halifax
America/Havana
America/Hermosillo
America/Indiana/Indianapolis
America/Indiana/Knox
America/Indiana/Marengo
America/Indiana/Petersburg
America/Indiana/Tell_City
America/Indiana/Vevay
America/Indiana/Vincennes
America/Indiana/Winamac
America/Indianapolis
America/Inuvik
America/Iqaluit
America/Jamaica
America/Jujuy
America/Juneau
America/Kentucky/Louisville
America/Kentucky/Monticello
America/Knox_IN
America/Kralendijk
America/La_Paz
America/Lima
America/Los_Angeles
America/Louisville
America/Lower_Princes
America/Maceio
America/Managua
America/Manaus
America/Marigot
America/Martinique
America/Matamoros
America/Mazatlan
America/Mendoza
America/Menominee
America/Merida
America/Metlakatla
America/Mexico_City
America/Miquelon
America/Moncton
America/Monterrey
America/Montevideo
America/Montreal
America/Montserrat
America/Nassau
America/New_York
America/Nipigon
America/Nome
America/Noronha
America/North_Dakota/Beulah
America/North_Dakota/Center
America/North_Dakota/New_Salem
America/Nuuk
America/Ojinaga
America/Panama
America/Pangnirtung
America/Paramaribo
America/Phoenix
America/Port-au-Prince
America/Port_of_Spain
America/Porto_Acre
America/Porto_Velho
America/Puerto_Rico
America/Punta_Arenas
America/Rainy_River
America/Rankin_Inlet
America/Recife
America/Regina
America/Resolute
America/Rio_Branco
America/Rosario
America/Santa_Isabel
America/Santarem
America/Santiago
America/Santo_Domingo
America/Sao_Paulo
America/Scoresbysund
America/Shiprock
America/Sitka
America/St_Barthelemy
America/St_Johns
America/St_Kitts
America/St_Lucia
America/St_Thomas
America/St_Vincent
America/Swift_Current
America/Tegucigalpa
America/Thule
America/Thunder_Bay
America/Tijuana
America/Toronto
America/Tortola
America/Vancouver
America/Virgin
America/Whitehorse
America/Winnipeg
America/Yakutat
America/Yellowknife
Antarctica/Casey
Antarctica/Davis
Antarctica/DumontDUrville
Antarctica/Macquarie
Antarctica/Mawson
Antarctica/McMurdo
Antarctica/Palmer
Antarctica/Rothera
Antarctica/South_Pole
Antarctica/Syowa
Antarctica/Troll
Antarctica/Vostok
Arctic/Longyearbyen
Asia/Aden
Asia/Almaty
Asia/Amman
Asia/Anadyr
Asia/Aqtau
Asia/Aqtobe
Asia/Ashgabat
Asia/Ashkhabad
Asia/Atyrau
Asia/Baghdad
Asia/Bahrain
Asia/Baku
Asia/Bangkok
Asia/Barnaul
Asia/Beirut
Asia/Bishkek
Asia/Brunei
Asia/Calcutta
Asia/Chita
Asia/Choibalsan
Asia/Chongqing
Asia/Chungking
Asia/Colombo
Asia/Dacca
Asia/Damascus
Asia/Dhaka
Asia/Dili
Asia/Dubai
Asia/Dushanbe
Asia/Famagusta
Asia/Gaza
Asia/Harbin
Asia/Hebron
Asia/Ho_Chi_Minh
Asia/Hong_Kong
Asia/Hovd
Asia/Irkutsk
Asia/Istanbul
Asia/Jakarta
Asia/Jayapura
Asia/Jerusalem
Asia/Kabul
Asia/Kamchatka
Asia/Karachi
Asia/Kashgar
Asia/Kathmandu
Asia/Katmandu
Asia/Khandyga
Asia/Kolkata
Asia/Krasnoyarsk
Asia/Kuala_Lumpur
Asia/Kuching
Asia/Kuwait
Asia/Macao
Asia/Macau
Asia/Magadan
Asia/Makassar
Asia/Manila
Asia/Muscat
Asia/Nicosia
Asia/Novokuznetsk
Asia/Novosibirsk
Asia/Omsk
Asia/Oral
Asia/Phnom_Penh
Asia/Pontianak
Asia/Pyongyang
Asia/Qatar
Asia/Qostanay
Asia/Qyzylorda
Asia/Rangoon
Asia/Riyadh
Asia/Saigon
Asia/Sakhalin
Asia/Samarkand
Asia/Seoul
Asia/Shanghai
Asia/Singapore
Asia/Srednekolymsk
Asia/Taipei
Asia/Tashkent
Asia/Tbilisi
Asia/Tehran
Asia/Tel_Aviv
Asia/Thimbu
Asia/Thimphu
Asia/Tokyo
Asia/Tomsk
Asia/Ujung_Pandang
Asia/Ulaanbaatar
Asia/Ulan_Bator
Asia/Urumqi
Asia/Ust-Nera
Asia/Vientiane
Asia/Vladivostok
Asia/Yakutsk
Asia/Yangon
Asia/Yekaterinburg
Asia/Yerevan
Atlantic/Azores
Atlantic/Bermuda
Atlantic/Canary
Atlantic/Cape_Verde
Atlantic/Faeroe
Atlantic/Faroe
Atlantic/Jan_Mayen
Atlantic/Madeira
Atlantic/Reykjavik
Atlantic/South_Georgia
Atlantic/St_Helena
Atlantic/Stanley
Australia/ACT
Australia/Adelaide
Australia/Brisbane
Australia/Broken_Hill
Australia/Canberra
Australia/Currie
Australia/Darwin
Australia/Eucla
Australia/Hobart
Australia/LHI
Australia/Lindeman
Australia/Lord_Howe
Australia/Melbourne
Australia/NSW
Australia/North
Australia/Perth
Australia/Queensland
Australia/South
Australia/Sydney
Australia/Tasmania
Australia/Victoria
Australia/West
Australia/Yancowinna
Brazil/Acre
Brazil/DeNoronha
Brazil/East
Brazil/West
CET
CST6CDT
Canada/Atlantic
Canada/Central
Canada/Eastern
Canada/Mountain
Canada/Newfoundland
Canada/Pacific
Canada/Saskatchewan
Canada/Yukon
Chile/Continental
Chile/EasterIsland
Cuba
EET
EST
EST5EDT
Egypt
Eire
Etc/GMT
Etc/GMT+0
Etc/GMT+1
Etc/GMT+10
Etc/GMT+11
Etc/GMT+12
Etc/GMT+2
Etc/GMT+3
Etc/GMT+4
Etc/GMT+5
Etc/GMT+6
Etc/GMT+7
Etc/GMT+8
Etc/GMT+9
Etc/GMT-0
Etc/GMT-1
Etc/GMT-10
Etc/GMT-11
Etc/GMT-12
Etc/GMT-13
Etc/GMT-14
Etc/GMT-2
Etc/GMT-3
Etc/GMT-4
Etc/GMT-5
Etc/GMT-6
Etc/GMT-7
Etc/GMT-8
Etc/GMT-9
Etc/GMT0
Etc/Greenwich
Etc/UCT
Etc/UTC
Etc/Universal
Etc/Zulu
Europe/Amsterdam
Europe/Andorra
Europe/Astrakhan
Europe/Athens
Europe/Belfast
Europe/Belgrade
Europe/Berlin
Europe/Bratislava
Europe/Brussels
Europe/Bucharest
Europe/Budapest
Europe/Busingen
Europe/Chisinau
Europe/Copenhagen
Europe/Dublin
Europe/Gibraltar
Europe/Guernsey
Europe/Helsinki
Europe/Isle_of_Man
Europe/Istanbul
Europe/Jersey
Europe/Kaliningrad
Europe/Kiev
Europe/Kirov
Europe/Kyiv
Europe/Lisbon
Europe/Ljubljana
Europe/London
Europe/Luxembourg
Europe/Madrid
Europe/Malta
Europe/Mariehamn
Europe/Minsk
Europe/Monaco
Europe/Moscow
Europe/Nicosia
Europe/Oslo
Europe/Paris
Europe/Podgorica
Europe/Prague
Europe/Riga
Europe/Rome
Europe/Samara
Europe/San_Marino
Europe/Sarajevo
Europe/Saratov
Europe/Simferopol
Europe/Skopje
Europe/Sofia
Europe/Stockholm
Europe/Tallinn
Europe/Tirane
Europe/Tiraspol
Europe/Ulyanovsk
Europe/Uzhgorod
Europe/Vaduz
Europe/Vatican
Europe/Vienna
Europe/Vilnius
Europe/Volgograd
Europe/Warsaw
Europe/Zagreb
Europe/Zaporozhye
Europe/Zurich
GB
GB-Eire
GMT
GMT+0
GMT-0
GMT0
Greenwich
HST
Hongkong
Iceland
Indian/Antananarivo
Indian/Chagos
Indian/Christmas
Indian/Cocos
Indian/Comoro
Indian/Kerguelen
Indian/Mahe
Indian/Maldives
Indian/Mauritius
Indian/Mayotte
Indian/Reunion
Iran
Israel
Jamaica
Japan
Kwajalein
Libya
MET
MST
MST7MDT
Mexico/BajaNorte
Mexico/BajaSur
Mexico/General
NZ
NZ-CHAT
Navajo
PRC
PST8PDT
Pacific/Apia
Pacific/Auckland
Pacific/Bougainville
Pacific/Chatham
Pacific/Chuuk
Pacific/Easter
Pacific/Efate
Pacific/Enderbury
Pacific/Fakaofo
Pacific/Fiji
Pacific/Funafuti
Pacific/Galapagos
Pacific/Gambier
Pacific/Guadalcanal
Pacific/Guam
Pacific/Honolulu
Pacific/Johnston
Pacific/Kanton
Pacific/Kiritimati
Pacific/Kosrae
Pacific/Kwajalein
Pacific/Majuro
Pacific/Marquesas
Pacific/Midway
Pacific/Nauru
Pacific/Niue
Pacific/Norfolk
Pacific/Noumea
Pacific/Pago_Pago
Pacific/Palau
Pacific/Pitcairn
Pacific/Pohnpei
Pacific/Ponape
Pacific/Port_Moresby
Pacific/Rarotonga
Pacific/Saipan
Pacific/Samoa
Pacific/Tahiti
Pacific/Tarawa
Pacific/Tongatapu
Pacific/Truk
Pacific/Wake
Pacific/Wallis
Pacific/Yap
Poland
Portugal
ROC
ROK
Singapore
Turkey
UCT
US/Alaska
US/Aleutian
US/Arizona
US/Central
US/East-Indiana
US/Eastern
US/Hawaii
US/Indiana-Starke
US/Michigan
US/Mountain
US/Pacific
US/Samoa
UTC
Universal
W-SU
WET
Zulu
""".strip().split("\n")

for timezone in _timezones:
    def make_reader_writer(prefix, timezone):
        @writer(f"{prefix}.{timezone}")
        def write_timezone(data, out):
            pass

        @reader(f"{prefix}.{timezone}")
        def read_timezone(data):
            import pytz
            return pytz.timezone(timezone)

    make_reader_writer("pytz.tzfile", timezone)

make_reader_writer("pytz", "UTC")



from safeserialize import dumps, loads

import scipy.sparse
import numpy as np

def test_scipy():
    np.random.seed(0)
    m = 20
    n = 10
    A = np.random.rand(m, n)
    A[np.random.rand(m, n) < 0.9] = 0

    matrices = {
        "bsr": scipy.sparse.bsr_matrix(A),
        "csr": scipy.sparse.csr_matrix(A),
        "csc": scipy.sparse.csc_matrix(A),
        "coo": scipy.sparse.coo_matrix(A),
    }

    data = dumps(matrices)

    loaded_matrices = loads(data)

    for name, expected_value in matrices.items():
        value = loaded_matrices[name]

        assert (value != expected_value).nnz == 0



from safeserialize import dumps, loads
import pytz
from safeserialize.types.pytz import _timezones

def test_pytz():
    for name in _timezones:
        timezone = pytz.timezone(name)
        serialized = dumps(timezone)
        deserialized = loads(serialized)
        assert timezone == deserialized



import uuid
from pathlib import Path
from decimal import Decimal
from datetime import datetime
from fractions import Fraction
from collections import deque, Counter, OrderedDict

from safeserialize import dumps, loads

def test_stdlib():
    a = datetime(2025, 8, 25, 22, 55, 47, 343776)
    b = datetime.fromtimestamp(0)

    data = {
        "fraction": Fraction(3, 4),
        "decimal": Decimal("3.14159265358979323846"),
        "datetime": a,
        "date": a.date(),
        "time": a.time(),
        "timedelta": a - b,
        "timedelta_negative": b - a,
        "uuid": uuid.uuid5(uuid.NAMESPACE_DNS, "example.com"),
        "path": Path("some/relative/path"),
        "deque": deque([1, 2, 3]),
        "counter": Counter("banana"),
        "ordered_dict": OrderedDict([("apple", 1), ("banana", 2)]),
    }

    serialized = dumps(data)

    deserialized = loads(serialized)

    assert data == deserialized, f"{data} != {deserialized}"



import uuid

from ...core import writer, reader, read_bytes, write_bytes

@writer("uuid.UUID")
def write_uuid(data, out):
    write_bytes(data.bytes, out)

@reader("uuid.UUID")
def read_uuid(f):
    return uuid.UUID(bytes=read_bytes(f))



from pathlib import Path

from ...core import writer, reader, read_str, write_str

"""
On Linux, "a\\b" is a valid file name,
not a directory "a" with a file "b".
We convert to "a/b" with as_posix().
"""

@writer("pathlib._local.PosixPath")
def write_local_posix_path(data, out):
    write_str(data.as_posix(), out)

@reader("pathlib._local.PosixPath")
def read_local_posix_path(f):
    return Path(read_str(f))

@writer("pathlib.PosixPath")
def write_posix_path(data, out):
    write_str(data.as_posix(), out)

@reader("pathlib.PosixPath")
def read_posix_path(f):
    return Path(read_str(f))

@writer("pathlib._local.WindowsPath")
def write_local_windows_path(data, out):
    write_str(data.as_posix(), out)

@reader("pathlib._local.WindowsPath")
def read_local_windows_path(f):
    return Path(read_str(f))

@writer("pathlib.WindowsPath")
def write_windows_path(data, out):
    write_str(data.as_posix(), out)

@reader("pathlib.WindowsPath")
def read_windows_path(f):
    return Path(read_str(f))



from safeserialize import dump, load, dumps, loads, dump_base64, load_base64
from safeserialize.core import num_bytes_signed_int

import tempfile

def test_num_bytes_signed_int():
    assert num_bytes_signed_int(1) == 1
    assert num_bytes_signed_int(-1) == 1
    assert num_bytes_signed_int(127) == 1
    assert num_bytes_signed_int(-128) == 1
    assert num_bytes_signed_int(128) == 2
    assert num_bytes_signed_int(-129) == 2
    assert num_bytes_signed_int(32767) == 2
    assert num_bytes_signed_int(32768) == 3
    assert num_bytes_signed_int(-32768) == 2
    assert num_bytes_signed_int(-32769) == 3
    assert num_bytes_signed_int(2147483647) == 4
    assert num_bytes_signed_int(-2147483648) == 4

def test_builtins():
    for x in range(-300, 300):
        data = dumps(x)
        y = loads(data)

        assert x == y, f"{x} != {y}"

    data = {
        1: [1, 2.0, 3, float("inf"), float("-inf")],
        3: [4, 5, 6],
        (1, 2): 3,
        frozenset([7, "foo", 9]): 4,
        123456789: 1 << 256,
        b"key": bytearray(b"value"),
        "float": 3.14159265358979323846,
        "true": True,
        "None": None,
        "list": [[], [[[]], []], [], [[], [[[]]]]],
        "set": {True, False, None, 1, 2, 3},
        "complex": 1 + 2j,
        "range": range(2, 10, -1),
        "slice": slice(None, 0.5, 123),
        "ellipsis": Ellipsis,
        "NotImplemented": NotImplemented,
    }

    serialized = dumps(data)

    deserialized = loads(serialized)

    assert data == deserialized, f"{data} != {deserialized}"

    with tempfile.NamedTemporaryFile(delete=True) as temp_file:
        dump(data, temp_file)
        temp_file.seek(0)
        deserialized = load(temp_file)

    assert data == deserialized, f"{data} != {deserialized}"

    assert data == load_base64(dump_base64(data))

def test_headerless():
    data = 1

    serialized = dumps(data, header=False)

    assert len(serialized) == 1

    deserialized = loads(serialized, header=False)

    assert data == deserialized

def test_constants():
    assert loads(dumps(True)) is True
    assert loads(dumps(False)) is False
    assert loads(dumps(None)) is None
    assert loads(dumps(...)) is ...
    assert loads(dumps(NotImplemented)) is NotImplemented



import io
import base64
import struct

from .constants import (
    builtins,
    VERSION,
    FILE_SIGNATURE,
    TYPE_BOOL_FALSE,
    TYPE_BOOL_TRUE,
    TYPE_INT1,
    TYPE_INT2,
    TYPE_INT4,
    TYPE_LARGE_INT,
    TYPE_CUSTOM,
    TYPE_INT_VALUE_0,
)

writers = {}
readers = {}

def writer(type_name):
    """Register a writer function for a type-name."""
    def decorator(func):
        assert type_name not in writers
        writers[type_name] = func
        return func
    return decorator

def reader(type_name):
    """Register a reader function for a type-name."""
    def decorator(func):
        assert type_name not in readers
        readers[type_name] = func
        if type_name in builtins:
            readers[builtins[type_name]] = func
        return func
    return decorator

class Serializer:
    def __init__(self, file, writers=None, header=True, version=VERSION):
        if writers is None:
            writers = {}

        if header:
            file.write(FILE_SIGNATURE)
            file.write(struct.pack("<I", version))

        self._file = file
        self._version = version
        self._writers = writers

    def write(self, data: bytes):
        """This write function mimics file.write. It does not serialize."""
        self._file.write(data)

    def tell(self):
        return self._file.tell()

class Deserializer:
    def __init__(self, file, readers=None, header=True, version=None):
        if readers is None:
            readers = {}

        self._file = file
        self._readers = readers

        if header:
            signature = self.read(len(FILE_SIGNATURE))

            if signature != FILE_SIGNATURE:
                raise ValueError(f"Invalid file signature {repr(signature)}")

            version, = struct.unpack("<I", self.read(4))
        else:
            if version is None:
                version = VERSION

        self._version = version

    def read(self, n):
        """Read bytes from underlying file. Does not deserialize.
        Throws ValueError if too few bytes were read."""
        data = self._file.read(n)

        if len(data) != n:
            raise ValueError(f"Expected {n} bytes, got {len(data)}")

        return data

def write(data, out):
    """Serialize data object and write it to output stream."""

    if data is False:
        out.write(bytes([TYPE_BOOL_FALSE]))
        return

    if data is True:
        out.write(bytes([TYPE_BOOL_TRUE]))
        return

    # Ints always include their type byte,
    # so they have to be handled differently from the other types.
    if isinstance(data, int):
        write_int(data, out)
        return

    cls = data.__class__
    data_type = f"{cls.__module__}.{cls.__name__}"

    out_writers = getattr(out, "_writers", {})

    if data_type not in out_writers:
        if data_type not in writers:
            raise TypeError(f"Writer not implemented for {type(data)}")
        else:
            writer = writers[data_type]
    else:
        writer = out_writers[data_type]

    type_id = builtins.get(data_type, TYPE_CUSTOM)

    out.write(bytes([type_id]))

    # Custom types get a type name
    if type_id == TYPE_CUSTOM:
        write_str(data_type, out)

    writer(data, out)

def read(f):
    """Read bytes from input stream and deserialize."""

    data_type, = f.read(1)

    f_readers = getattr(f, "_readers", {})

    if data_type in f_readers:
        return f_readers[data_type](f)

    elif data_type in readers:
        return readers[data_type](f)

    elif data_type == TYPE_CUSTOM:
        name = read_str(f)

        if name not in f_readers:
            if name not in readers:
                raise TypeError(f"Reader not implemented for {repr(name)}")
            else:
                reader = readers[name]
        else:
            reader = f_readers[name]

        return reader(f)
    else:
        raise TypeError(f"Reader not implemented for {repr(data_type)}")

def dump(obj, file, writers=None, header=True):
    """Serialize object to a file."""
    serializer = Serializer(file, writers, header)
    write(obj, serializer)

def dumps(obj, writers=None, header=True):
    """Serialize object to bytes."""
    out = io.BytesIO()
    serializer = Serializer(out, writers, header)
    write(obj, serializer)
    return out.getvalue()

def load(file, readers=None, header=True, version=None):
    """Deserialize object from a file."""
    deserializer = Deserializer(file, readers, header, version)
    return read(deserializer)

def loads(data, readers=None, header=True, version=None):
    """Deserialize object from a file."""
    deserializer = Deserializer(io.BytesIO(data), readers, header, version)
    return read(deserializer)

def dump_base64(data: bytes) -> str:
    """Like dumps, but encodes result as base64."""
    return base64.b64encode(dumps(data)).decode("ascii")

def load_base64(text: str) -> bytes:
    """Like loads, but decodes input from base64 beforehand."""
    return loads(base64.b64decode(text.encode("ascii")))

def read_int(f):
    result = read(f)

    if not isinstance(result, int):
        raise ValueError(f"Expected int, got {type(result)}")

    return result

def num_bytes_signed_int(data):
    bits = data.bit_length() or 1
    n = (bits + 7) // 8
    sign_bit = 1 << (8 * n - 1)

    # Need extra byte if sign bit is set
    if (data >= 0 and data >= sign_bit) or (data < 0 and data < -sign_bit):
        n += 1

    return n

def write_int(data, out):
    # Ints are encoded with different type IDs based on their size.
    if 0 <= data < 10:
        out.write(bytes([TYPE_INT_VALUE_0 + data]))
    elif -128 <= data <= 127:
        out.write(bytes([TYPE_INT1]))
        out.write(struct.pack("<b", data))
    elif -32768 <= data <= 32767:
        out.write(bytes([TYPE_INT2]))
        out.write(struct.pack("<h", data))
    elif -2147483648 <= data <= 2147483647:
        out.write(bytes([TYPE_INT4]))
        out.write(struct.pack("<i", data))
    else:
        out.write(bytes([TYPE_LARGE_INT]))
        num_bytes = num_bytes_signed_int(data)
        out.write(struct.pack("<Q", num_bytes))
        out.write(data.to_bytes(num_bytes, byteorder="little", signed=True))

for i in range(10):
    def make_reader(i):
        return lambda _: i
    reader(TYPE_INT_VALUE_0 + i)(make_reader(i))

@reader(TYPE_INT1)
def read_int1(f):
    return struct.unpack("<b", f.read(1))[0]

@reader(TYPE_INT2)
def read_int2(f):
    return struct.unpack("<h", f.read(2))[0]

@reader(TYPE_INT4)
def read_int4(f):
    return struct.unpack("<i", f.read(4))[0]

@reader(TYPE_LARGE_INT)
def read_large_int(f):
    num_bytes, = struct.unpack("<Q", f.read(8))
    return int.from_bytes(f.read(num_bytes), byteorder="little", signed=True)

@reader(TYPE_BOOL_FALSE)
def read_false(f):
    return False

@reader(TYPE_BOOL_TRUE)
def read_true(f):
    return True

@writer("builtins.list")
def write_list(data, out):
    write_int(len(data), out)
    for value in data:
        write(value, out)

@reader("builtins.list")
def read_list(f):
    length = read_int(f)
    return [read(f) for _ in range(length)]

@writer("builtins.dict")
def write_dict(data, out):
    write_int(len(data), out)
    for key, value in data.items():
        write(key, out)
        write(value, out)

@reader("builtins.dict")
def read_dict(f):
    length = read_int(f)
    return {read(f): read(f) for _ in range(length)}

@writer("builtins.tuple")
def write_tuple(data, out):
    write_int(len(data), out)
    for value in data:
        write(value, out)

@reader("builtins.tuple")
def read_tuple(f):
    return tuple(read_list(f))

@writer("builtins.set")
def write_set(data, out):
    write_int(len(data), out)
    for value in data:
        write(value, out)

@reader("builtins.set")
def read_set(f):
    return set(read_list(f))

@writer("builtins.frozenset")
def write_frozenset(data, out):
    write_int(len(data), out)
    for value in data:
        write(value, out)

@reader("builtins.frozenset")
def read_frozenset(f):
    return frozenset(read_list(f))

@writer("builtins.bytes")
def write_bytes(data, out):
    write_int(len(data), out)
    out.write(data)

@reader("builtins.bytes")
def read_bytes(f):
    length = read_int(f)
    return f.read(length)

@writer("builtins.bytearray")
def write_bytearray(data, out):
    write_int(len(data), out)
    out.write(data)

@reader("builtins.bytearray")
def read_bytearray(f):
    length = read_int(f)
    return bytearray(f.read(length))

@writer("builtins.str")
def write_str(data, out):
    write_bytes(data.encode("utf-8"), out)

@reader("builtins.str")
def read_str(f):
    return read_bytes(f).decode("utf-8")

@writer("builtins.float")
def write_float(data, out):
    out.write(struct.pack("<d", data))

@reader("builtins.float")
def read_float(f):
    return struct.unpack("<d", f.read(8))[0]

@writer("builtins.bool")
def write_bool(data, out):
    out.write(bytes([1 if data else 0]))

@reader("builtins.bool")
def read_bool(f):
    return bool(f.read(1)[0])

@writer("builtins.NoneType")
def write_none(data, out):
    pass

@reader("builtins.NoneType")
def read_none(f):
    return None

@writer("builtins.complex")
def write_complex(data, out):
    out.write(struct.pack("<dd", data.real, data.imag))

@reader("builtins.complex")
def read_complex(f):
    real, imag = struct.unpack("<dd", f.read(2 * 8))
    return complex(real, imag)

@writer("builtins.range")
def write_range(data, out):
    write_int(data.start, out)
    write_int(data.stop, out)
    write_int(data.step, out)

@reader("builtins.range")
def read_range(f):
    start = read_int(f)
    stop = read_int(f)
    step = read_int(f)
    return range(start, stop, step)

@writer("builtins.slice")
def write_slice(data, out):
    write(data.start, out)
    write(data.stop, out)
    write(data.step, out)

@reader("builtins.slice")
def read_slice(f):
    start = read(f)
    stop = read(f)
    step = read(f)
    return slice(start, stop, step)

@writer("builtins.ellipsis")
def write_ellipsis(data, out):
    pass

@reader("builtins.ellipsis")
def read_ellipsis(f):
    return Ellipsis

@writer("builtins.NotImplementedType")
def write_not_implemented(data, out):
    pass

@reader("builtins.NotImplementedType")
def read_not_implemented(f):
    return NotImplemented



from safeserialize import read, write, loads, dumps

class Person:
    def __init__(self, name, age):
        self.name = name
        self.age = age

    def __eq__(self, other):
        return self.name == other.name and self.age == other.age

def write_person(person, out):
    write(person.name, out)
    write(person.age, out)

def read_person(f):
    name = read(f)
    age = read(f)
    return Person(name, age)

def test_custom():
    people = [
        Person("Bilbo", 111),
        Person("Gandalf", 2000),
    ]

    # Name must match module hierarchy + class name
    name = "tests.test_custom_type.Person"

    serialized_data = dumps(people, writers={name: write_person})

    loaded_people = loads(serialized_data, readers={name: read_person})

    assert people == loaded_people



import struct
import ast

from ..core import writer, reader, write_list, read_list

_allowed_dtypes = {
    "bool",
    "uint8", "uint16", "uint32", "uint64",
    "int8", "int16", "int32", "int64",
    "float16", "float32", "float64",
    "complex64", "complex128",
    "datetime64[ns]", "timedelta64[ns]",
    "object",
}

_descriptors = {
    "|b1",
    "|i1", "<i2", "<i4", "<i8",
    "|u1", "<u2", "<u4", "<u8",
    "<f2", "<f4", "<f8",
    "<c8", "<c16",
    "<m8[ns]", "<M8[ns]",
    "|O",
}

@writer("numpy.ndarray")
def write_ndarray(data, out, alignment=64):
    dtype = str(data.dtype)

    if dtype not in _allowed_dtypes:
        raise NotImplementedError(f"NumPy dtype {data.dtype} not supported")

    if alignment is not None:
        # Pad with leading zeros until aligned.
        # This is useful for some assembly instructions,
        # which require that kind of alignment.
        # Also, NumPy tries to align data to 64 byte boundary,
        # so we try to preserve that.
        pos = out.tell()

        if pos % alignment != 0:
            out.write(b"\x00" * (alignment - pos % alignment))

    magic = b"\x93NUMPY"

    version = bytes([1, 0])

    (_, descr), = data.dtype.descr

    header = "{"
    header += f"'descr': '{descr}', "
    header += "'fortran_order': False, "
    header += f"'shape': {data.shape}, "
    header += "}"
    header = header.encode("ascii")

    # magic + version + length + header (+ padding) + newline
    length = len(magic) + 2 + 2 + len(header) + 1

    padding = 0
    if (length + padding) % 64 != 0:
        padding = 64 - (length + padding) % 64

    header = b"".join([
        magic,
        version,
        struct.pack("<H", len(header) + padding + 1),
        header,
        b" " * padding,
        b"\n",
    ])

    assert len(header) % 64 == 0

    out.write(header)

    if dtype == "object":
        write_list(data.ravel(order="C"), out)
    else:
        out.write(data.tobytes())

@reader("numpy.ndarray")
def read_ndarray(f):
    import numpy as np

    magic = b"\x93NUMPY\1\0"

    buf = b""
    # Consume leading padding zeros
    for _ in range(10):
        buf += f.read(len(magic))

        idx = buf.find(magic[:1])

        # If we got the first byte of magic,
        # we can compute how much more magic we need to read.
        if idx != -1:
            # How much magic do we have already?
            have = len(buf) - idx
            remaining = len(magic) - have
            buf += f.read(remaining)
            break

    assert buf.lstrip(b"\0") == magic

    header_len, = struct.unpack("<H", f.read(2))

    assert (10 + header_len) % 64 == 0

    dict_data = f.read(header_len)

    assert dict_data.endswith(b"\n")

    d = ast.literal_eval(dict_data.decode("utf-8"))

    assert list(d) == ["descr", "fortran_order", "shape"]
    assert d["fortran_order"] is False

    descr = d["descr"]
    shape = d["shape"]

    assert descr in _descriptors, f"NumPy dtype {repr(descr)} not implemented"

    if descr == "|O":
        objects = read_list(f)
        return np.array(objects, dtype=object).reshape(shape)
    else:
        itemsize = int(descr[2:4].rstrip("["))

        num_bytes = np.prod(shape) * itemsize

        buf = f.read(num_bytes)

        return np.frombuffer(buf, dtype=descr).reshape(shape).copy()

@writer("numpy.bool")
def write_bool(data, out):
    out.write(struct.pack("<?", data))

@reader("numpy.bool")
def read_bool(f):
    import numpy as np
    return np.bool_(struct.unpack("<?", f.read(1))[0])

@writer("numpy.int32")
def write_int32(data, out):
    out.write(struct.pack("<i", data))

@reader("numpy.int32")
def read_int32(f):
    import numpy as np
    return np.int32(struct.unpack("<i", f.read(4))[0])

@writer("numpy.int64")
def write_int64(data, out):
    out.write(struct.pack("<q", data))

@reader("numpy.int64")
def read_int64(f):
    import numpy as np
    return np.int64(struct.unpack("<q", f.read(8))[0])

@writer("numpy.float16")
def write_float16(data, out):
    out.write(struct.pack("<e", data))

@reader("numpy.float16")
def read_float16(f):
    import numpy as np
    return np.float16(struct.unpack("<e", f.read(2))[0])

@writer("numpy.float32")
def write_float32(data, out):
    out.write(struct.pack("<f", data))

@reader("numpy.float32")
def read_float32(f):
    import numpy as np
    return np.float32(struct.unpack("<f", f.read(4))[0])

@writer("numpy.float64")
def write_float64(data, out):
    out.write(struct.pack("<d", data))

@reader("numpy.float64")
def read_float64(f):
    import numpy as np
    return np.float64(struct.unpack("<d", f.read(8))[0])

@writer("numpy.complex64")
def write_complex64(data, out):
    out.write(struct.pack("<ff", data.real, data.imag))

@reader("numpy.complex64")
def read_complex64(f):
    import numpy as np
    real, imag = struct.unpack("<ff", f.read(8))
    return np.complex64(real + 1j * imag)

@writer("numpy.complex128")
def write_complex128(data, out):
    out.write(struct.pack("<dd", data.real, data.imag))

@reader("numpy.complex128")
def read_complex128(f):
    import numpy as np
    real, imag = struct.unpack("<dd", f.read(16))
    return np.complex128(real + 1j * imag)

@writer("numpy.int8")
def write_int8(data, out):
    out.write(struct.pack("<b", data))

@reader("numpy.int8")
def read_int8(f):
    import numpy as np
    return np.int8(struct.unpack("<b", f.read(1))[0])

@writer("numpy.int16")
def write_int16(data, out):
    out.write(struct.pack("<h", data))

@reader("numpy.int16")
def read_int16(f):
    import numpy as np
    return np.int16(struct.unpack("<h", f.read(2))[0])

@writer("numpy.uint8")
def write_uint8(data, out):
    out.write(struct.pack("<B", data))

@reader("numpy.uint8")
def read_uint8(f):
    import numpy as np
    return np.uint8(struct.unpack("<B", f.read(1))[0])

@writer("numpy.uint16")
def write_uint16(data, out):
    out.write(struct.pack("<H", data))

@reader("numpy.uint16")
def read_uint16(f):
    import numpy as np
    return np.uint16(struct.unpack("<H", f.read(2))[0])

@writer("numpy.uint32")
def write_uint32(data, out):
    out.write(struct.pack("<I", data))

@reader("numpy.uint32")
def read_uint32(f):
    import numpy as np
    return np.uint32(struct.unpack("<I", f.read(4))[0])

@writer("numpy.uint64")
def write_uint64(data, out):
    out.write(struct.pack("<Q", data))

@reader("numpy.uint64")
def read_uint64(f):
    import numpy as np
    return np.uint64(struct.unpack("<Q", f.read(8))[0])

dtypes = [
    ("Int8DType", "int8"),
    ("Int16DType", "int16"),
    ("Int32DType", "int32"),
    ("Int64DType", "int64"),
    ("UInt8DType", "uint8"),
    ("UInt16DType", "uint16"),
    ("UInt32DType", "uint32"),
    ("UInt64DType", "uint64"),
    ("BoolDType", "bool"),
    ("Float16DType", "float16"),
    ("Float32DType", "float32"),
    ("Float64DType", "float64"),
    ("Complex64DType", "complex64"),
    ("Complex128DType", "complex128"),
    ("DateTime64DType", "datetime64[ns]"),
    ("TimeDelta64DType", "timedelta64[ns]"),
    ("ObjectDType", "object"),
]

for type_name, dtype_str in dtypes:
    def make_dtype_reader_writer(type_name, dtype_str):
        @writer(f"numpy.dtypes.{type_name}")
        def writer_func(data, out):
            pass

        @reader(f"numpy.dtypes.{type_name}")
        def reader_func(f):
            import numpy as np
            return np.dtype(dtype_str)

    make_dtype_reader_writer(type_name, dtype_str)



from safeserialize import dumps, loads
import pandas as pd
import numpy as np
import random
from safeserialize.types.numpy import _allowed_dtypes as numpy_dtypes

# Functions to test whether data == loads(dumps(data))
# for pd.Series, pd.DataFrame and pd.Index
def roundtrip_series(s):
    serialized_data = dumps(s)
    deserialized_series = loads(serialized_data)
    pd.testing.assert_series_equal(s, deserialized_series)

def roundtrip_df(df):
    serialized_data = dumps(df)
    deserialized_df = loads(serialized_data)
    pd.testing.assert_frame_equal(df, deserialized_df)

def roundtrip_index(index):
    serialized_data = dumps(index)
    deserialized_index = loads(serialized_data)
    pd.testing.assert_index_equal(index, deserialized_index)

def test_pandas():
    # Test various data types
    a = pd.Series([1, 2, None, 4], dtype="Int64", name="int_nullable")
    b = pd.Series([3.14, np.nan, 2.71828], dtype="Float32", name="float32")
    c = pd.Series([True, False, None], dtype="boolean", name="bool_nullable")
    d = pd.Series(["foo", None, "bar"], dtype="string", name="string")
    data = ["1678-01-01", "2262-04-11"]
    series = pd.Series(data, name="datetime_series")
    e = pd.to_datetime(series, utc=False)
    e.name = "datetime"
    data = ["1 day", None, "1 sec", "02:00:00"]
    series = pd.Series(data, name="timedelta_series")
    f = pd.to_timedelta(series)
    f.name = "timedelta"
    g = pd.Series([{"one": 1}, None, [{1, 2}]], dtype="object", name="object")
    h = pd.Series([1, None, 3, 4], dtype="UInt32", name="uint_nullable")
    i = pd.Series(np.arange(5), name="arange")
    assert i.dtype == "int64"
    j = pd.Series(np.linspace(0, 1, 11), name="linspace")
    assert j.dtype == "float64"

    series = [a, a, b, c, d, e, f, g, h, i, j]

    # Test individual series
    for s in series:
        roundtrip_series(s)

    # Test series combined into one DataFrame
    df = pd.concat(series, axis=1)

    roundtrip_df(df)

    # DataFrame with duplicate column names
    df = pd.concat([a, a, b], axis=1)

    roundtrip_df(df)

    # DataFrame with renamed columns
    df = pd.concat([b, d], axis=1)

    df = df.rename(columns={"string": "d", "float32": "b"})

    # Check names
    for column, series in df.items():
        assert series.name == column

    roundtrip_df(df)

def test_categories():
    for categories in [
        [1, 2, 3, 4],
        [1, 2, 3, 4, None],
        ["red", "green", "blue", 123, None],
        [1.0, 2.0, 3.0],
        [1.0, 2.0, 3.0, None],
    ]:
        values = [random.choice(categories) for _ in range(50)]
        s = pd.Series(values, name="Adelheid").astype("category")

        roundtrip_series(s)

    dtype = pd.CategoricalDtype(["n", "b", "a"], ordered=True)
    s = pd.Series(list("banana"), dtype=dtype, name="Berthold")

    roundtrip_series(s)

    index = pd.CategoricalIndex([1, 2, 3], categories=[1, 2, 3], name="Conrad")

    roundtrip_index(index)

    series = pd.Series([1, 2, 3], name="Dietlinde", index=index)

    roundtrip_series(series)

def test_numpy_dtypes():
    for dtype in numpy_dtypes:
        data = [0, 1, 0, 1, 0, 0, 0, 1, 1, 1]
        s = pd.Series(data, dtype=dtype, name=f"numpy_{dtype}")
        roundtrip_series(s)

def test_datetime():
    df = pd.DataFrame({
        "year": [2025, 2026],
        "month": [1, 2],
        "day": [1, 2]})

    series = pd.to_datetime(df)
    series.name = "Eberhard"

    roundtrip_series(series)

    t = pd.to_timedelta("1 days 01:02:03.00004")
    series = pd.Series([t], name="Frieda")

    roundtrip_series(series)

    # Test datetime with timezone. Timezone-aware objects
    # should always be checked with and without timezone
    # because internally, Pandas often stores timestamps with
    # NumPy, which is not timezone aware, and the conversion is
    # easy to get wrong.
    start = pd.to_datetime("1/1/2025").tz_localize("Europe/Berlin")
    end = pd.to_datetime("12/31/2025").tz_localize("Europe/Berlin")
    index = pd.date_range(start=start, end=end, name="Gerhardt")

    series = pd.Series(index, name="Gisela")

    data = [pd.Timestamp("1/1/1970").tz_localize("Europe/Berlin")]
    df = pd.DataFrame({"s": data})
    roundtrip_df(df)

    index = pd.DatetimeIndex(data, name="Gunther")

    series = pd.Series(data, index=index, name="Hedwig")

    roundtrip_series(series)

    roundtrip_df(pd.DataFrame({"s": series}, index=index))

    index = pd.CategoricalIndex(data, name="Heinrich")

    series = pd.Series(data, index=index, name="Irmgard")

    roundtrip_series(series)

    df = pd.DataFrame({"s": series}, index=index)

    roundtrip_df(df)

def test_datetime_index_naive():
    index = pd.to_datetime(["2023-01-01", "2023-02-01", "2023-03-01"])
    index.name = "Karl"
    roundtrip_index(index)

def test_timedelta_index():
    index = pd.timedelta_range(
        start="1 day",
        end="5 days",
        freq="D",
        name="Hugo")

    roundtrip_index(index)
    series = pd.Series([1, 2, 3, 4, 5], index=index)
    roundtrip_series(series)
    df = pd.DataFrame({"values": [1, 2, 3, 4, 5]}, index=index)
    roundtrip_df(df)

def test_interval_index():
    index = pd.interval_range(
        start=0,
        end=20,
        freq=3,
        closed="both",
        name="Mathilde")

    serialized_data = dumps(index)
    deserialized_index = loads(serialized_data)
    pd.testing.assert_index_equal(index, deserialized_index)

    data = [1, 2, 3, 4, 5, 6]
    series = pd.Series(data, index=index, name="Oswald")
    roundtrip_series(series)

    df = pd.DataFrame({"values": data}, index=index)
    roundtrip_df(df)

    for i in range(6):
        interval = df.index[i]
        assert df.loc[interval, "values"] == data[i]

    test_interval = pd.Interval(0, 3, closed="both")
    assert test_interval in df.index

def test_interval_index_with_datetime():
    start = pd.to_datetime("1/1/2025").tz_localize("Europe/Berlin")
    end = pd.to_datetime("12/31/2025").tz_localize("Europe/Berlin")
    index = pd.interval_range(
        start=start,
        end=end,
        freq="2D",
        closed="both",
        name="Reinhilde")

    roundtrip_index(index)

    data = [start + pd.Timedelta(days=i * 2) for i in range(len(index))]

    roundtrip_series(pd.Series(data, index=index, name="Siegfried"))

    df = pd.DataFrame({"dates": data}, index=index)

    roundtrip_df(df)

def test_period_index():
    index = pd.period_range(
        start="2000-01-01",
        end="2001-01-01",
        freq="M",
        name="Egon")

    roundtrip_index(index)

    series = pd.Series(range(len(index)), index=index, name="Hugo")
    roundtrip_series(series)

    df = pd.DataFrame({"values": range(len(index))}, index=index)
    roundtrip_df(df)

def test_multi_index():
    arrays = [[1, 1, 2, 2], ["red", "blue", "red", "blue"]]
    names = ["number", "color"]
    index = pd.MultiIndex.from_arrays(arrays, sortorder=1, names=names)
    roundtrip_index(index)

    series = pd.Series(np.random.randn(4), index=index, name="Manfred")
    roundtrip_series(series)

    df = pd.DataFrame({"values": np.random.randn(4)}, index=index)
    roundtrip_df(df)

def test_categorical_index_advanced():
    # With None in data
    categories = ["a", "b", "c"]
    data = ["a", "b", "a", None, "c"]
    index = pd.CategoricalIndex(data, categories=categories, name="with_none")
    roundtrip_index(index)

    # Ordered with custom order
    categories = ["low", "medium", "high"]
    data = ["medium", "high", "low", "medium"]
    index = pd.CategoricalIndex(
        data,
        categories=categories,
        ordered=True,
        name="ordered_custom")

    roundtrip_index(index)

    # Unused categories
    categories = ["apple", "banana", "cherry", "date"]
    data = ["apple", "cherry", "apple"]
    index = pd.CategoricalIndex(
        data,
        categories=categories,
        name="unused_categories")

    roundtrip_index(index)

    # Datetime with timezone
    tz = "America/New_York"
    dates = ["2023-01-01", "2023-01-02", "2023-01-03"]
    categories = pd.to_datetime(dates).tz_localize(tz)
    data = [categories[0], categories[2], categories[0]]
    index = pd.CategoricalIndex(
        data,
        categories=categories,
        name="datetime_tz")

    roundtrip_index(index)

    # Empty data
    index = pd.CategoricalIndex(
        [],
        categories=["x", "y", "z"],
        name="empty_with_categories")

    roundtrip_index(index)



from collections import deque, Counter, OrderedDict

from ...core import writer, reader, read, write, read_int, write_int

# TODO: defaultdict for some reasonable default types (list, int, ...)?

@writer("collections.deque")
def write_deque(data, out):
    write_int(len(data), out)
    for value in data:
        write(value, out)

@reader("collections.deque")
def read_deque(f):
    length = read_int(f)
    return deque(read(f) for _ in range(length))

@writer("collections.Counter")
def write_counter(data, out):
    write_int(len(data), out)
    for key, value in data.items():
        write(key, out)
        write(value, out)

@reader("collections.Counter")
def read_counter(f):
    length = read_int(f)
    c = Counter()
    for _ in range(length):
        key = read(f)
        value = read(f)
        c[key] = value
    return c

@writer("collections.OrderedDict")
def write_ordered_dict(data, out):
    write_int(len(data), out)
    for key, value in data.items():
        write(key, out)
        write(value, out)

@reader("collections.OrderedDict")
def read_ordered_dict(f):
    length = read_int(f)
    d = OrderedDict()
    for _ in range(length):
        key = read(f)
        value = read(f)
        d[key] = value
    return d



import torch
from safeserialize import dumps, loads

def test_float():
    x = torch.rand(2, 3, 4)

    roundtrip(x)

def test_cuda():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    x = torch.rand(2, 3, 4)

    x = x.to(device)

    roundtrip(x)

def test_long():
    x = torch.arange(5)

    roundtrip(x)

def test_half():
    x = torch.rand(5).half()

    roundtrip(x)

def roundtrip(x):
    s = dumps(x)
    y = loads(s)
    assert torch.equal(x, y)

def test_transposed():
    A = torch.rand(2, 3)

    roundtrip(A)



from datetime import datetime, date, time, timedelta
import struct

from ...core import writer, reader, read_str, write_str

@writer("datetime.datetime")
def write_datetime(data, out):
    write_str(data.isoformat(), out)

@reader("datetime.datetime")
def read_datetime(f):
    return datetime.fromisoformat(read_str(f))

@writer("datetime.date")
def write_date(data, out):
    write_str(data.isoformat(), out)

@reader("datetime.date")
def read_date(f):
    return date.fromisoformat(read_str(f))

@writer("datetime.time")
def write_time(data, out):
    write_str(data.isoformat(), out)

@reader("datetime.time")
def read_time(f):
    return time.fromisoformat(read_str(f))

@writer("datetime.timedelta")
def write_timedelta(data, out):
    # Use nanoseconds in case we need more precision in the future
    nanoseconds = data.microseconds * 1000
    seconds = data.seconds
    days = data.days
    assert -999999999 <= days <= 999999999
    assert 0 <= seconds <= 999999999
    assert 0 <= nanoseconds <= 999999999
    out.write(struct.pack("<iII", days, seconds, nanoseconds))

@reader("datetime.timedelta")
def read_timedelta(f):
    days, seconds, nanoseconds = struct.unpack("<iII", f.read(3 * 4))
    microseconds = nanoseconds // 1000
    return timedelta(days=days, seconds=seconds, microseconds=microseconds)



from ..core import writer, reader, write, read

_allowed_dtypes = {
    "bool",
    "uint8", "uint16", "uint32", "uint64",
    "int8", "int16", "int32", "int64",
    "float16", "float32", "float64",
    "complex64", "complex128",
}

VERSION = 1

@writer("torch.Tensor")
def write_tensor(data, out):
    write(VERSION, out)
    device = data.device
    write(str(device), out)
    data_np = data.detach().cpu().numpy()
    assert str(data_np.dtype) in _allowed_dtypes
    write(data_np, out)

@reader("torch.Tensor")
def read_tensor(f):
    version = read(f)
    assert version == VERSION
    device = read(f)
    import torch
    data = read(f)
    assert str(data.dtype) in _allowed_dtypes
    return torch.from_numpy(data).to(device)



from fractions import Fraction

from ...core import writer, reader, read_int, write_int

@writer("fractions.Fraction")
def write_fraction(data, out):
    write_int(data.numerator, out)
    write_int(data.denominator, out)

@reader("fractions.Fraction")
def read_fraction(f):
    numerator = read_int(f)
    denominator = read_int(f)
    return Fraction(numerator, denominator)



from safeserialize import dumps, loads
import io
import tempfile
import numpy as np
from safeserialize.types.numpy import write_ndarray, read_ndarray

def test_numpy_serialization():
    np.random.seed(0)

    x = np.array([1, 2, 3])

    data = {
        "3d array of doubles": np.random.rand(1, 2, 3),
        "1d array of ints": np.arange(5),
        "bool array": np.random.rand(5) < 0.5,
        "object array": np.array([{"one": 1}, None, {2: "two"}, {3, 4}]),
        "x.sum()": x.sum(),
        "bool": np.bool_(True),
        "int8": np.int8(-128),
        "int16": np.int16(-32768),
        "int32": np.int32(-2147483648),
        "int64": np.int64(-9223372036854775808),
        "uint8": np.uint8(255),
        "uint16": np.uint16(65535),
        "uint32": np.uint32(4294967295),
        "uint64": np.uint64(18446744073709551615),
        "float16": np.float16(2.71828),
        "float32": np.float32(3.14),
        "float64": np.float64(3.141592653589793),
        "complex64": np.complex64(1 + 2j),
        "complex128": np.complex128(1 + 2j),
    }

    serialized = dumps(data)

    deserialized = loads(serialized)

    for key, expected_value in data.items():
        value = deserialized[key]
        assert np.array_equal(value, expected_value)

def test_write_ndarray():
    # Verify write_ndarray against NumPy np.save
    np.random.seed(0)

    a = np.random.rand(3, 4, 5)

    with tempfile.NamedTemporaryFile(delete=True, suffix=".npy") as tmp:
        np.save(tmp.name, a)
        tmp.seek(0)
        expected = tmp.read()

    f = io.BytesIO()
    write_ndarray(a, f)
    f.seek(0)
    data = f.read()

    assert data == expected

def test_read_ndarray():
    np.random.seed(0)

    a = np.random.rand(12, 13)

    with tempfile.NamedTemporaryFile(delete=True, suffix=".npy") as tmp:
        np.save(tmp.name, a)
        tmp.seek(0)

        b = read_ndarray(tmp)

    assert np.array_equal(a, b)
    assert np.array_equal(a, b)

def test_transposed():
    A = np.random.rand(2, 3)

    assert np.array_equal(A.T, loads(dumps(A.T)))



from safeserialize import write, read
from ..core import writer, reader

VERSION = 1

@writer("scipy.sparse._bsr.bsr_matrix")
def write_bsr_matrix(data, out):
    m, n = data.shape
    write(VERSION, out)
    write(m, out)
    write(n, out)
    write(data.indptr, out)
    write(data.indices, out)
    write(data.data, out)

@reader("scipy.sparse._bsr.bsr_matrix")
def read_bsr_matrix(f):
    version = read(f)
    assert version == VERSION
    m = read(f)
    n = read(f)
    indptr = read(f)
    indices = read(f)
    data = read(f)
    import scipy.sparse
    return scipy.sparse.bsr_matrix((data, indices, indptr), shape=(m, n))

@writer("scipy.sparse._csr.csr_matrix")
def write_csr_matrix(data, out):
    m, n = data.shape
    write(VERSION, out)
    write(m, out)
    write(n, out)
    write(data.indptr, out)
    write(data.indices, out)
    write(data.data, out)

@reader("scipy.sparse._csr.csr_matrix")
def read_csr_matrix(f):
    version = read(f)
    assert version == VERSION
    m = read(f)
    n = read(f)
    indptr = read(f)
    indices = read(f)
    data = read(f)
    import scipy.sparse
    return scipy.sparse.csr_matrix((data, indices, indptr), shape=(m, n))

@writer("scipy.sparse._csc.csc_matrix")
def write_csc_matrix(data, out):
    m, n = data.shape
    write(VERSION, out)
    write(m, out)
    write(n, out)
    write(data.indptr, out)
    write(data.indices, out)
    write(data.data, out)

@reader("scipy.sparse._csc.csc_matrix")
def read_csc_matrix(f):
    version = read(f)
    assert version == VERSION
    m = read(f)
    n = read(f)
    indptr = read(f)
    indices = read(f)
    data = read(f)
    import scipy.sparse
    return scipy.sparse.csc_matrix((data, indices, indptr), shape=(m, n))

@writer("scipy.sparse._coo.coo_matrix")
def write_coo_matrix(data, out):
    m, n = data.shape
    write(VERSION, out)
    write(m, out)
    write(n, out)
    row, col = data.coords
    write(row, out)
    write(col, out)
    write(data.data, out)

@reader("scipy.sparse._coo.coo_matrix")
def read_coo_matrix(f):
    version = read(f)
    assert version == VERSION
    m = read(f)
    n = read(f)
    row = read(f)
    col = read(f)
    data = read(f)
    import scipy.sparse
    return scipy.sparse.coo_matrix((data, (row, col)), shape=(m, n))



from .core import (
    dump, load,
    dumps, loads,
    dump_base64, load_base64,
    Serializer, Deserializer,
    write, read,
)

from . import types

# Import all type modules to register their readers and writers
from .types import (
    numpy,
    scipy,
    pandas,
    torch,
    pytz,
)
from .types.stdlib import (
    datetime,
    decimal,
    fractions,
    uuid,
    pathlib,
    collections,
)
